from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any, Generic, Type, TypeVar

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from pydiverse.transform.core.expressions.translator import TypedValue
    from pydiverse.transform.core.table_impl import AbstractTableImpl


def generate_col_uuid():
    return uuid.uuid1()


ImplT = TypeVar("ImplT", bound="AbstractTableImpl")
LiteralT = TypeVar("LiteralT")


class Column(Generic[ImplT]):
    __slots__ = ("name", "table", "dtype", "uuid")

    def __init__(self, name: str, table: ImplT, dtype: str, uuid: uuid.UUID = None):
        self.name = name
        self.table = table
        self.dtype = dtype
        self.uuid = uuid or generate_col_uuid()

    def __repr__(self):
        return f"<{self.table.name}.{self.name}({self.dtype})>"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name and self.uuid == other.uuid

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.uuid)


class LambdaColumn:
    """Anonymous Column

    A lambda column is a column without an associated table or UUID. This means
    that it can be used to reference columns in the same pipe as it was created.

    Example:
      The following fails because `table.a` gets referenced before it gets created.
        table >> mutate(a = table.x) >> mutate(b = table.a)
      Instead you can use a lambda column to achieve this:
        table >> mutate(a = table.x) >> mutate(b = λ.a)
    """

    __slots__ = "name"

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f"<λ.{self.name}>"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(("λ", self.name))


class LiteralColumn(Generic[LiteralT]):
    __slots__ = ("typed_value", "expr", "backend")

    def __init__(
        self,
        typed_value: TypedValue[LiteralT],
        expr: Any,
        backend: type[AbstractTableImpl],
    ):
        self.typed_value = typed_value
        self.expr = expr
        self.backend = backend

    def __repr__(self):
        return f"<Lit: {self.expr} ({self.typed_value.dtype})>"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.typed_value == other.typed_value
            and self.expr == other.expr
            and self.backend == other.backend
        )

    def __ne__(self, other):
        return not self.__eq__(other)



from __future__ import annotations

from .core import Binary, ElementWise, Unary

__all__ = [
    "Equal",
    "NotEqual",
    "Less",
    "LessEqual",
    "Greater",
    "GreaterEqual",
    "And",
    "RAnd",
    "Or",
    "ROr",
    "Xor",
    "RXor",
    "Invert",
]


#### Comparison Operators ####


class Comparison(ElementWise, Binary):
    signatures = [
        "int, int -> bool",
        "int, float -> bool",
        "float, int -> bool",
        "float, float -> bool",
        "str, str -> bool",
        "bool, bool -> bool",
    ]

    def validate_signature(self, signature):
        assert signature.rtype == "bool"
        super().validate_signature(signature)


class Equal(Comparison):
    name = "__eq__"


class NotEqual(Comparison):
    name = "__ne__"


class Less(Comparison):
    name = "__lt__"


class LessEqual(Comparison):
    name = "__le__"


class Greater(Comparison):
    name = "__gt__"


class GreaterEqual(Comparison):
    name = "__ge__"


#### Boolean Operators ####


class BooleanBinary(ElementWise, Binary):
    signatures = [
        "bool, bool -> bool",
    ]

    def validate_signature(self, signature):
        assert signature.args == ("bool", "bool")
        assert signature.rtype == "bool"
        super().validate_signature(signature)


class And(BooleanBinary):
    name = "__and__"


class RAnd(BooleanBinary):
    name = "__rand__"


class Or(BooleanBinary):
    name = "__or__"


class ROr(BooleanBinary):
    name = "__ror__"


class Xor(BooleanBinary):
    name = "__xor__"


class RXor(BooleanBinary):
    name = "__rxor__"


class Invert(ElementWise, Unary):
    name = "__invert__"
    signatures = [
        "bool -> bool",
    ]



from __future__ import annotations

from .core import Nullary, Window

__all__ = [
    "Shift",
    "RowNumber",
]


class Shift(Window):
    name = "shift"
    signatures = [
        "T, int -> T",
        "T, int, T -> T",
    ]


class RowNumber(Window, Nullary):
    name = "row_number"
    signatures = [
        "-> int",
    ]



from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeVar
from collections.abc import Iterable

from pydiverse.transform.core import column
from pydiverse.transform.core.expressions import expressions
from pydiverse.transform.core.ops import Operator, OPType, registry

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from pydiverse.transform.core.table_impl import AbstractTableImpl

ImplT = TypeVar("ImplT", bound="AbstractTableImpl")
T = TypeVar("T")


# Basic container to store value and associated type metadata
@dataclass
class TypedValue(Generic[T]):
    value: T
    dtype: str
    ftype: OPType = dataclasses.field(default=OPType.EWISE)

    def __iter__(self):
        return iter((self.value, self.dtype))


class Translator(Generic[T]):
    def translate(self, expr, **kwargs) -> T:
        """Translate an expression recursively."""
        try:
            return bottom_up_replace(expr, lambda e: self._translate(e, **kwargs))
        except Exception as e:
            raise ValueError(
                "An exception occured while trying to translate the expression"
                f" '{expr}':\n{e}"
            ) from e

    def _translate(self, expr, **kwargs) -> T:
        """Translate an expression non recursively."""
        raise NotImplementedError


class DelegatingTranslator(Generic[T], Translator[T]):
    """
    Translator that dispatches to different translate functions based on
    the type of the expression.
    """

    def __init__(self, operator_registry: registry.OperatorRegistry):
        self.operator_registry = operator_registry

    def translate(self, expr, **kwargs):
        """Translate an expression recursively."""
        try:
            return self._translate(expr, **kwargs)
        except Exception as e:
            raise ValueError(
                "An exception occured while trying to translate the expression"
                f" '{expr}':\n{e}"
            ) from e

    def _translate(self, expr, accept_literal_col=True, **kwargs):
        if isinstance(expr, column.Column):
            return self._translate_col(expr, **kwargs)

        if isinstance(expr, column.LiteralColumn):
            if accept_literal_col:
                return self._translate_literal_col(expr, **kwargs)
            else:
                raise ValueError("Literal columns aren't allowed in this context.")

        if isinstance(expr, expressions.FunctionCall):
            operator = self.operator_registry.get_operator(expr.name)
            op_args, op_kwargs, context_kwargs = self.__translate_function_arguments(
                expr, operator, **kwargs
            )

            if op_kwargs:
                raise NotImplementedError

            signature = tuple(arg.dtype for arg in op_args)
            implementation = self.operator_registry.get_implementation(
                expr.name, signature
            )

            return self._translate_function(
                expr, implementation, op_args, context_kwargs, **kwargs
            )

        if literal_result := self._translate_literal(expr, **kwargs):
            return literal_result

        raise NotImplementedError(
            f"Couldn't find a way to translate object of type {type(expr)} with value"
            f" {expr}."
        )

    def _translate_col(self, expr: column.Column, **kwargs) -> T:
        raise NotImplementedError

    def _translate_literal_col(self, expr: column.LiteralColumn, **kwargs) -> T:
        raise NotImplementedError

    def _translate_function(
        self,
        expr: expressions.FunctionCall,
        implementation: registry.TypedOperatorImpl,
        op_args: Iterable[T],
        context_kwargs: dict[str, Any],
        **kwargs,
    ) -> T:
        raise NotImplementedError

    def _translate_literal(self, expr, **kwargs) -> T:
        raise NotImplementedError

    def __translate_function_arguments(
        self, expr: expressions.FunctionCall, operator: Operator, **kwargs
    ):
        op_args = [self._translate(arg, **kwargs) for arg in expr.args]
        op_kwargs = {}
        context_kwargs = {}

        for k, v in expr.kwargs.items():
            if k in operator.context_kwargs:
                context_kwargs[k] = v
            else:
                op_kwargs[k] = self._translate(v, **kwargs)

        return op_args, op_kwargs, context_kwargs


def bottom_up_replace(expr, replace):
    def transform(expr):
        if isinstance(expr, expressions.FunctionCall):
            f = expressions.FunctionCall(
                expr.name,
                *(transform(arg) for arg in expr.args),
                **{k: transform(v) for k, v in expr.kwargs.items()},
            )
            return replace(f)
        else:
            return replace(expr)

    return transform(expr)



from __future__ import annotations

import functools
import operator
import uuid
import warnings
from functools import reduce
from typing import Callable, Dict

import sqlalchemy
from sqlalchemy import sql

from pydiverse.transform.core import ops
from pydiverse.transform.core.column import Column, LiteralColumn
from pydiverse.transform.core.expressions import SymbolicExpression, iterate_over_expr
from pydiverse.transform.core.expressions.translator import TypedValue
from pydiverse.transform.core.ops import OPType
from pydiverse.transform.core.table_impl import ColumnMetaData
from pydiverse.transform.core.util import OrderingDescriptor, translate_ordering

from .lazy_table import JoinDescriptor, LazyTableImpl


class SQLTableImpl(LazyTableImpl):
    """SQL backend

    Attributes:
        tbl: The underlying SQLAlchemy table object.
        engine: The SQLAlchemy engine.
        sql_columns: A dict mapping from uuids to SQLAlchemy column objects
            (only those contained in `tbl`).

        alignment_hash: A hash value that allows checking if two tables are
            'aligned'. In the case of SQL this means that two tables NUST NOT
            share the same alignment hash unless they were derived from the
            same table(s) and are guaranteed to have the same number of columns
            in the same order. In other words: Two tables MUST only have the
            same alignment hash if a literal column derived from one of them
            can be used by the other table and produces the same result.
    """

    def __init__(
        self, engine, table, dataset=None, _dtype_hints: dict[str, str] = None
    ):
        self.engine = (
            sqlalchemy.create_engine(engine) if isinstance(engine, str) else engine
        )
        tbl = self._create_table(table, self.engine)
        # backend = self.engine.url.get_backend_name()

        columns = {
            col.name: Column(
                name=col.name,
                table=self,
                dtype=self._get_dtype(col, hints=_dtype_hints),
            )
            for col in tbl.columns
        }

        self.replace_tbl(tbl, columns)

        name = self.tbl.name
        if engine.dialect.name == "bigquery":
            spl = self.tbl.name.split(".")
            if len(spl) == 2:
                name = spl[1]

        super().__init__(name=name, columns=columns)

    def is_aligned_with(self, col: Column | LiteralColumn) -> bool:
        if isinstance(col, Column):
            if not isinstance(col.table, type(self)):
                return False
            return col.table.alignment_hash == self.alignment_hash

        if isinstance(col, LiteralColumn):
            return all(
                self.is_aligned_with(atom)
                for atom in iterate_over_expr(col.expr, expand_literal_col=True)
                if isinstance(atom, Column)
            )

        raise ValueError

    @classmethod
    def _html_repr_expr(cls, expr):
        if isinstance(expr, sqlalchemy.sql.expression.ColumnElement):
            return str(expr.compile(compile_kwargs={"literal_binds": True}))
        return super()._html_repr_expr(expr)

    @staticmethod
    def _create_table(tbl, engine=None):
        """Return a sqlalchemy.Table

        Arguments:
            tbl: a sqlalchemy.Table or string of form 'table_name' or 'schema_name.table_name'.
        """
        if isinstance(tbl, sqlalchemy.sql.selectable.FromClause):
            return tbl

        if not isinstance(tbl, str):
            raise ValueError(
                "tbl must be a sqlalchemy Table or string, but was %s" % type(tbl)
            )

        spl = tbl.split(".") if "." in tbl else [None, tbl]
        schema = spl[0]
        table_name = ".".join(spl[1:])

        # TODO: pybigquery uses schema to mean project_id, so we cannot use
        #     siuba's classic breakdown "{schema}.{table_name}". Basically
        #     pybigquery uses "{schema=project_id}.{dataset_dot_table_name}" in its internal
        #     logic. An important side effect is that bigquery errors for
        #     `dataset`.`table`, but not `dataset.table`.
        if engine and engine.dialect.name == "bigquery":
            table_name = tbl
            schema = None

        return sqlalchemy.Table(
            table_name,
            sqlalchemy.MetaData(),
            schema=schema,
            autoload_with=engine,
        )

    @staticmethod
    def _get_dtype(col: sqlalchemy.Column, hints: dict[str, str] = None) -> str:
        """Determine the dtype of a column.

        :param col: The sqlalchemy column object.
        :param hints: In some situations sqlalchemy can't determine the dtype of
            a column. Instead of throwing an exception we can use these type
            hints as a fallback.
        :return: Appropriate dtype string.
        """

        try:
            pytype = col.type.python_type
            if pytype == int:
                return "int"
            if pytype == str:
                return "str"
            if pytype == bool:
                return "bool"
            if pytype == float:
                return "float"
            raise NotImplementedError(f"Invalid type {col.type}.")
        except NotImplementedError as e:
            if hints is not None:
                if dtype := hints.get(col.name):
                    return dtype
            raise e

    def replace_tbl(self, new_tbl, columns: dict[str:Column]):
        if isinstance(new_tbl, sql.Select):
            # noinspection PyNoneFunctionAssignment
            new_tbl = new_tbl.subquery()

        self.tbl = new_tbl
        self.alignment_hash = generate_alignment_hash()

        self.sql_columns = {
            col.uuid: self.tbl.columns[col.name] for col in columns.values()
        }  # from uuid to sqlalchemy column

        if hasattr(self, "cols"):
            # TODO: Clean up... This feels a bit hacky
            for col in columns.values():
                self.cols[col.uuid] = ColumnMetaData.from_expr(col.uuid, col, self)
        if hasattr(self, "intrinsic_grouped_by"):
            self.intrinsic_grouped_by.clear()

        self.joins: list[JoinDescriptor] = []
        self.wheres: list[SymbolicExpression] = []
        self.having: list[SymbolicExpression] = []
        self.order_bys: list[OrderingDescriptor] = []
        self.limit_offset: tuple[int, int] = None

    def build_select(self) -> sql.Select:
        # Validate current state
        if len(self.selects) == 0:
            raise ValueError("Can't execute a SQL query without any SELECT statements.")

        # Start building query
        select = self.tbl.select()

        # `select_from` is required if no table is explicitly referenced
        # inside the SELECT. e.g. `SELECT COUNT(*) AS count`
        select = select.select_from(self.tbl)

        # FROM
        if self.joins:
            for join in self.joins:
                compiled, _ = self.compiler.translate(join.on)
                on = compiled(self.sql_columns)

                select = select.join(
                    join.right.tbl,
                    onclause=on,
                    isouter=join.how != "inner",
                    full=join.how == "outer",
                )

        # WHERE
        if self.wheres:
            # Combine wheres using ands
            combined_where = functools.reduce(
                operator.and_, map(SymbolicExpression, self.wheres)
            )._
            compiled, where_dtype = self.compiler.translate(combined_where)
            assert where_dtype == "bool"
            where = compiled(self.sql_columns)
            select = select.where(where)

        # GROUP BY
        if self.intrinsic_grouped_by:
            compiled_gb, group_by_dtypes = zip(
                *(
                    self.compiler.translate(group_by)
                    for group_by in self.intrinsic_grouped_by
                )
            )
            group_bys = (compiled(self.sql_columns) for compiled in compiled_gb)
            select = select.group_by(*group_bys)

        # HAVING
        if self.having:
            # Combine havings using ands
            combined_having = functools.reduce(
                operator.and_, map(SymbolicExpression, self.having)
            )._
            compiled, having_dtype = self.compiler.translate(combined_having)
            assert having_dtype == "bool"
            having = compiled(self.sql_columns)
            select = select.having(having)

        # LIMIT / OFFSET
        if self.limit_offset is not None:
            limit, offset = self.limit_offset
            select = select.limit(limit).offset(offset)

        # SELECT
        # Convert self.selects to SQLAlchemy Expressions
        s = []
        for name, uuid in self.selected_cols():
            sql_col = self.cols[uuid].compiled(self.sql_columns)
            if not isinstance(sql_col, sql.ColumnElement):
                sql_col = sql.literal(sql_col)
            s.append(sql_col.label(name))

        select = select.with_only_columns(*s)

        # ORDER BY
        if self.order_bys:
            o = []
            for o_by in self.order_bys:
                compiled, _ = self.compiler.translate(o_by.order)
                col = compiled(self.sql_columns)
                col = col.asc() if o_by.asc else col.desc()
                col = col.nullsfirst() if o_by.nulls_first else col.nullslast()
                o.append(col)
            select = select.order_by(*o)

        return select

    #### Verb Operations ####

    def preverb_hook(self, verb: str, *args, **kwargs) -> None:
        def has_any_ftype_cols(ftypes: OPType | tuple[OPType, ...], cols: Iterable):
            if isinstance(ftypes, OPType):
                ftypes = (ftypes,)
            return any(
                self.cols[c.uuid].ftype in ftypes
                for v in cols
                for c in iterate_over_expr(self.resolve_lambda_cols(v))
                if isinstance(c, Column)
            )

        requires_subquery = False

        if self.limit_offset is not None:
            # The LIMIT / TOP clause is executed at the very end of the query.
            # This means we must create a subquery for any verb that modifies
            # the rows.
            if verb in (
                "join",
                "filter",
                "arrange",
                "group_by",
                "summarise",
            ):
                requires_subquery = True

        if verb == "mutate":
            # Window functions can't be nested, thus a subquery is required
            requires_subquery |= has_any_ftype_cols(OPType.WINDOW, kwargs.values())
        elif verb == "filter":
            # Window functions aren't allowed in where clause
            requires_subquery |= has_any_ftype_cols(OPType.WINDOW, args)
        elif verb == "summarise":
            # The result of the aggregate is always ordered according to the
            # grouping columns. We must clear the order_bys so that the order
            # is consistent with eager execution. We can do this because aggregate
            # functions are independent of the order.
            self.order_bys.clear()

            # If the grouping level is different from the grouping level of the
            # tbl object, or if on of the input columns is a window or aggregate
            # function, we must make a subquery.
            requires_subquery |= (
                bool(self.intrinsic_grouped_by)
                and self.grouped_by != self.intrinsic_grouped_by
            )
            requires_subquery |= has_any_ftype_cols(
                (OPType.AGGREGATE, OPType.WINDOW), kwargs.values()
            )

        # TODO: It would be nice if this could be done without having to select all columns.
        #       As a potential challenge for the hackathon I propose a mean of even creating the subqueries lazyly.
        #       This means that we could perform some kind of query optimization before submitting the actual query.
        #       Eg: Instead of selecting all possible columns, only select those that actually get used.
        if requires_subquery:
            columns = {
                name: self.cols[uuid].as_column(name, self)
                for name, uuid in self.named_cols.fwd.items()
            }

            original_selects = self.selects.copy()
            self.selects |= columns.keys()
            subquery = self.build_select()

            self.replace_tbl(subquery, columns)
            self.selects = original_selects

    def alias(self, name=None):
        if name is None:
            suffix = format(uuid.uuid1().int % 0x7FFFFFFF, "X")
            name = f"{self.name}_{suffix}"

        # TODO: If the table has not been modified, a simple `.alias()` would produce nicer queries.
        subquery = self.build_select().subquery(name=name)
        # In some situations sqlalchemy fails to determine the datatype of a column.
        # To circumvent this, we can pass on the information we know.
        dtype_hints = {
            name: self.cols[self.named_cols.fwd[name]].dtype for name in self.selects
        }
        return self.__class__(self.engine, subquery, _dtype_hints=dtype_hints)

    def collect(self):
        select = self.build_select()
        with self.engine.connect() as conn:
            try:
                # TODO: check for which pandas versions this is needed:
                # Temporary fix for pandas bug (https://github.com/pandas-dev/pandas/issues/35484)
                # Taken from siuba
                from pandas.io import sql as _pd_sql

                class _FixedSqlDatabase(_pd_sql.SQLDatabase):
                    def execute(self, *args, **kwargs):
                        return self.connectable.execute(*args, **kwargs)

                sql_db = _FixedSqlDatabase(conn)
                result = sql_db.read_sql(select).convert_dtypes()
            except AttributeError:
                import pandas as pd

                result = pd.read_sql_query(select, con=conn)

        # Add metadata
        result.attrs["name"] = self.name
        return result

    def build_query(self) -> str:
        query = self.build_select()
        return str(
            query.compile(
                dialect=self.engine.dialect, compile_kwargs={"literal_binds": True}
            )
        )

    def join(self, right, on, how, *, validate=None):
        self.alignment_hash = generate_alignment_hash()

        # If right has joins already, merging them becomes extremely difficult
        # This is because the ON clauses could contain NULL checks in which case
        # the joins aren't associative anymore.
        if right.joins:
            raise ValueError(
                "Can't automatically combine joins if the right side already contains a"
                " JOIN clause."
            )

        if right.limit_offset is not None:
            raise ValueError(
                "The right table can't be sliced when performing a join."
                " Wrap the right side in a subquery to fix this."
            )

        # TODO: Handle GROUP BY and SELECTS on left / right side

        # Combine the WHERE clauses
        if how == "inner":
            # Inner Join: The WHERES can be combined
            self.wheres.extend(right.wheres)
        elif how == "left":
            # WHERES from right must go into the ON clause
            on = reduce(operator.and_, (on, *right.wheres))
        elif how == "outer":
            # For outer joins, the WHERE clause can't easily be merged.
            # The best solution for now is to move them into a subquery.
            if self.wheres:
                raise ValueError(
                    "Filters can't precede outer joins. Wrap the left side in a"
                    " subquery to fix this."
                )
            if right.wheres:
                raise ValueError(
                    "Filters can't precede outer joins. Wrap the right side in a"
                    " subquery to fix this."
                )

        if validate is not None:
            warnings.warn("SQL table backend ignores join validation argument.")

        descriptor = JoinDescriptor(right, on, how)
        self.joins.append(descriptor)

        self.sql_columns.update(right.sql_columns)

    def filter(self, *args):
        self.alignment_hash = generate_alignment_hash()

        if self.intrinsic_grouped_by:
            for arg in args:
                # If a condition involves only grouping columns, it can be
                # moved into the wheres instead of the havings.
                only_grouping_cols = all(
                    col in self.intrinsic_grouped_by
                    for col in iterate_over_expr(arg, expand_literal_col=True)
                    if isinstance(col, Column)
                )

                if only_grouping_cols:
                    self.wheres.append(arg)
                else:
                    self.having.append(arg)
        else:
            self.wheres.extend(args)

    def arrange(self, ordering):
        self.alignment_hash = generate_alignment_hash()
        self.order_bys = ordering + self.order_bys

    def summarise(self, **kwargs):
        self.alignment_hash = generate_alignment_hash()

    def slice_head(self, n: int, offset: int):
        if self.limit_offset is None:
            self.limit_offset = (n, offset)
        else:
            old_n, old_o = self.limit_offset
            self.limit_offset = (min(abs(old_n - offset), n), old_o + offset)

    #### EXPRESSIONS ####

    class ExpressionCompiler(
        LazyTableImpl.ExpressionCompiler[
            "SQLTableImpl",
            TypedValue[
                Callable[[dict[uuid.UUID, sqlalchemy.Column]], sql.ColumnElement]
            ],
        ]
    ):
        def _translate_col(self, expr, **kwargs):
            # Can either be a base SQL column, or a reference to an expression
            if expr.uuid in self.backend.sql_columns:

                def sql_col(cols):
                    return cols[expr.uuid]

                return TypedValue(sql_col, expr.dtype, OPType.EWISE)

            col = self.backend.cols[expr.uuid]
            return TypedValue(col.compiled, col.dtype, col.ftype)

        def _translate_literal_col(self, expr, **kwargs):
            if not self.backend.is_aligned_with(expr):
                raise ValueError(
                    "Literal column isn't aligned with this table. "
                    f"Literal Column: {expr}"
                )

            def sql_col(cols):
                return expr.typed_value.value

            return TypedValue(sql_col, expr.typed_value.dtype, expr.typed_value.ftype)

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, verb=None, **kwargs
        ):
            def value(cols):
                return implementation(*(arg.value(cols) for arg in op_args))

            operator = implementation.operator

            if operator.ftype == OPType.AGGREGATE and verb == "mutate":
                # Aggregate function in mutate verb -> window function
                over_value = self.over_clause(
                    value, implementation.operator, context_kwargs
                )
                ftype = self.backend._get_op_ftype(
                    op_args, operator, OPType.WINDOW, strict=True
                )
                return TypedValue(over_value, implementation.rtype, ftype)

            elif operator.ftype == OPType.WINDOW:
                if verb != "mutate":
                    raise ValueError(
                        "Window function are only allowed inside a mutate."
                    )

                over_value = self.over_clause(
                    value, implementation.operator, context_kwargs
                )
                ftype = self.backend._get_op_ftype(op_args, operator, strict=True)
                return TypedValue(over_value, implementation.rtype, ftype)

            else:
                ftype = self.backend._get_op_ftype(op_args, operator, strict=True)
                return TypedValue(value, implementation.rtype, ftype)

        def over_clause(
            self, value: Callable, operator: ops.Operator, context_kwargs: dict
        ):
            if operator.ftype not in (OPType.AGGREGATE, OPType.WINDOW):
                raise ValueError

            wants_order_by = operator.ftype == OPType.WINDOW

            # PARTITION BY
            compiled_pb = tuple(
                self.translate(group_by).value for group_by in self.backend.grouped_by
            )

            # ORDER BY
            def order_by_clause_generator(ordering: OrderingDescriptor):
                compiled, _ = self.translate(ordering.order)

                def clause(*args, **kwargs):
                    col = compiled(*args, **kwargs)
                    col = col.asc() if ordering.asc else col.desc()
                    col = col.nullsfirst() if ordering.nulls_first else col.nullslast()
                    return col

                return clause

            if wants_order_by:
                arrange = context_kwargs.get("arrange")
                if not arrange:
                    raise TypeError("Missing 'arrange' argument.")

                ordering = translate_ordering(self.backend, arrange)
                compiled_ob = [order_by_clause_generator(o_by) for o_by in ordering]

            # New value callable
            def over_value(*args, **kwargs):
                pb = sql.expression.ClauseList(
                    *(compiled(*args, **kwargs) for compiled in compiled_pb)
                )
                ob = (
                    sql.expression.ClauseList(
                        *(compiled(*args, **kwargs) for compiled in compiled_ob)
                    )
                    if wants_order_by
                    else None
                )

                v = value(*args, **kwargs)
                return v.over(
                    partition_by=pb,
                    order_by=ob,
                )

            return over_value

    class AlignedExpressionEvaluator(
        LazyTableImpl.AlignedExpressionEvaluator[TypedValue[sql.ColumnElement]]
    ):
        def translate(self, expr, check_alignment=True, **kwargs):
            if check_alignment:
                alignment_hashes = {
                    col.table.alignment_hash
                    for col in iterate_over_expr(expr, expand_literal_col=True)
                    if isinstance(col, Column)
                }
                if len(alignment_hashes) >= 2:
                    raise ValueError(
                        "Expression contains columns from different tables that aren't"
                        " aligned."
                    )

            return super().translate(expr, check_alignment=check_alignment, **kwargs)

        def _translate_col(self, expr, **kwargs):
            backend = expr.table
            if expr.uuid in backend.sql_columns:
                sql_col = backend.sql_columns[expr.uuid]
                return TypedValue(sql_col, expr.dtype)

            col = backend.cols[expr.uuid]
            return TypedValue(col.compiled(backend.sql_columns), col.dtype, col.ftype)

        def _translate_literal_col(self, expr, **kwargs):
            assert issubclass(expr.backend, SQLTableImpl)
            return expr.typed_value

        def _translate_function(
            self, expr, implementation, op_args, context_kwargs, **kwargs
        ):
            # Aggregate function -> window function
            value = implementation(*(arg.value for arg in op_args))
            operator = implementation.operator
            override_ftype = (
                OPType.WINDOW if operator.ftype == OPType.AGGREGATE else None
            )
            ftype = SQLTableImpl._get_op_ftype(
                op_args, operator, override_ftype, strict=True
            )

            if operator.ftype == OPType.AGGREGATE:
                value = value.over()
            if operator.ftype == OPType.WINDOW:
                raise NotImplementedError("How to handle window functions?")

            return TypedValue(value, implementation.rtype, ftype)


def generate_alignment_hash():
    # It should be possible to have an alternative hash value that
    # is a bit more lenient -> If the same set of operations get applied
    # to a table in two different orders that produce the same table
    # object, their hash could also be equal.
    return uuid.uuid1()


#### BACKEND SPECIFIC OPERATORS ################################################


from sqlalchemy import func as sqlfunc

with SQLTableImpl.op(ops.FloorDiv(), check_super=False) as op:

    @op.auto
    def _floordiv(lhs, rhs):
        return sql.cast(lhs / rhs, sqlalchemy.types.Integer())


with SQLTableImpl.op(ops.RFloorDiv(), check_super=False) as op:

    @op.auto
    def _rfloordiv(rhs, lhs):
        return _floordiv(lhs, rhs)


with SQLTableImpl.op(ops.Abs()) as op:

    @op.auto
    def _abs(x):
        return sqlfunc.ABS(x)


with SQLTableImpl.op(ops.Round()) as op:

    @op.auto
    def _round(x, decimals=0):
        # TODO: Don't round integers with decimals >= 0
        if decimals >= 0:
            return sqlfunc.round(x, decimals)
        # For some reason SQLite doesn't like negative decimals values
        return sqlfunc.round(x / (10**-decimals)) * (10**-decimals)


with SQLTableImpl.op(ops.Strip()) as op:

    @op.auto
    def _strip(x):
        return sqlfunc.TRIM(x)


#### Summarising Functions ####


with SQLTableImpl.op(ops.Mean()) as op:

    @op.auto
    def _mean(x):
        return sqlfunc.AVG(x)


with SQLTableImpl.op(ops.Min()) as op:

    @op.auto
    def _min(x):
        return sqlfunc.MIN(x)


with SQLTableImpl.op(ops.Max()) as op:

    @op.auto
    def _max(x):
        return sqlfunc.MAX(x)


with SQLTableImpl.op(ops.Sum()) as op:

    @op.auto
    def _sum(x):
        return sqlfunc.SUM(x)


with SQLTableImpl.op(ops.StringJoin()) as op:

    @op.auto
    def _join(x, sep: str):
        return sqlfunc.GROUP_CONCAT(x, sep)


with SQLTableImpl.op(ops.Count()) as op:

    @op.auto
    def _count(x=None):
        if x is None:
            # Get the number of rows
            return sqlfunc.COUNT()
        else:
            # Count non null values
            return sqlfunc.COUNT(x)


#### Window Functions ####


with SQLTableImpl.op(ops.Shift()) as op:

    @op.auto
    def _shift(x, by, empty_value=None):
        if by == 0:
            return x
        if by > 0:
            return sqlfunc.LAG(x, by, empty_value)
        if by < 0:
            return sqlfunc.LEAD(x, -by, empty_value)
        raise Exception


with SQLTableImpl.op(ops.RowNumber()) as op:

    @op.auto
    def _row_number():
        return sqlfunc.ROW_NUMBER()



from __future__ import annotations

from . import symbolic_expressions


class FunctionCall:
    """
    AST node to represent a function / operator call.
    """

    def __init__(self, name: str, *args, **kwargs):
        # Unwrap all symbolic expressions in the input
        args = symbolic_expressions.unwrap_symbolic_expressions(args)
        kwargs = symbolic_expressions.unwrap_symbolic_expressions(kwargs)

        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        args = [repr(e) for e in self.args] + [
            f"{k}={repr(v)}" for k, v in self.kwargs.items()
        ]
        return f'{self.name}({", ".join(args)})'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.name, self.args, tuple(self.kwargs.items())))

    def iter_children(self):
        yield from self.args



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from pydiverse.transform.core.table_impl import AbstractTableImpl

ImplT = TypeVar("ImplT", bound=AbstractTableImpl)


@dataclass
class JoinDescriptor(Generic[ImplT]):
    __slots__ = ("right", "on", "how")

    right: ImplT
    on: Any
    how: str


class LazyTableImpl(AbstractTableImpl):
    """Base class for lazy backends."""

    pass



from __future__ import annotations

import pandas as pd

from pydiverse.transform.core import Table
from pydiverse.transform.core.verbs import collect, show_query


def assert_equal(left, right, check_dtype=False):
    l = left >> collect() if isinstance(left, Table) else left
    r = right >> collect() if isinstance(right, Table) else right

    try:
        pd.testing.assert_frame_equal(l, r, check_dtype=check_dtype)
    except AssertionError as e:
        print("First dataframe:")
        print(l)
        if isinstance(left, Table):
            left >> show_query()
        print()
        print("Second dataframe:")
        print(r)
        if isinstance(right, Table):
            right >> show_query()
        raise e



from __future__ import annotations

import dataclasses
import functools
import inspect
import itertools
import typing
from functools import partial

if typing.TYPE_CHECKING:
    from pydiverse.transform.core.ops import Operator, OperatorExtension


class OperatorImpl:
    """
    Internal type to store the implementation of an operator and all associated
    metadata.

    If the function (`impl`) provided as the underlying parameter has keyword
    only arguments that start with an underscore, they get added to the
    `self.internal_kwargs` list.
    """

    id = itertools.count()

    def __init__(
        self,
        operator: Operator,
        impl: typing.Callable,
        signature: OperatorSignature,
    ):
        self.operator = operator
        self.impl = impl
        self.signature = signature
        self.variants: dict[str, typing.Callable] = {}

        self.__id = next(OperatorImpl.id)

        # Inspect impl signature to get internal kwargs
        self.internal_kwargs = []

        try:
            impl_signature = inspect.signature(impl)
            for name, param in impl_signature.parameters.items():
                if param.kind == inspect.Parameter.KEYWORD_ONLY and name.startswith(
                    "_"
                ):
                    self.internal_kwargs.append(name)
        except (TypeError, ValueError):
            pass

        # Calculate Ordering Key
        # - Find match with the least number templates in the signature
        # - From those take the one with the least number of different templates
        # - From those take the one where the first template appears latest
        # - From those take the one that isn't a vararg or has the most arguments.
        # - From those take the one that was defined first

        num_templates = 0
        templates_set = set()
        first_template_index = len(signature.args)
        for i, dtype in enumerate(signature.args):
            if is_template_type(dtype):
                num_templates += 1
                templates_set.add(dtype)
                first_template_index = min(first_template_index, i)
        num_different_templates = len(templates_set)
        is_vararg = int(self.signature.is_vararg)

        self._precedence = (
            num_templates,
            num_different_templates,
            -first_template_index,
            is_vararg,
            -len(signature.args),
            self.__id,
        )

    def add_variant(self, name: str, impl: typing.Callable):
        if name in self.variants:
            raise ValueError(
                f"Already added a variant with name '{name}'"
                f" to operator {self.operator}."
            )
        self.variants[name] = impl


@dataclasses.dataclass
class TypedOperatorImpl:
    """
    Operator Implementation with a non-templated return type.
    Unlike `OperatorImpl`, this class is intended to be the return type of
    the OperatorRegistry.
    """

    operator: Operator
    impl: OperatorImpl
    rtype: str

    @classmethod
    def from_operator_impl(cls, impl: OperatorImpl, rtype: str):
        return cls(
            operator=impl.operator,
            impl=impl,
            rtype=rtype,
        )

    def __call__(self, *args, **kwargs):
        return self.impl.impl(*args, **self.__clean_kwargs(kwargs))

    def get_variant(self, name: str) -> typing.Callable | None:
        variant = self.impl.variants.get(name)
        if variant is None:
            return None

        @functools.wraps(variant)
        def variant_wrapper(*args, **kwargs):
            return variant(*args, **self.__clean_kwargs(kwargs))

        return variant_wrapper

    def __clean_kwargs(self, kwargs):
        return {
            k: v
            for k, v in kwargs.items()
            if not k.startswith("_") or k in self.impl.internal_kwargs
        }


class OperatorRegistry:
    # It only makes sense to define some dunder methods.
    # These are the ones which can be registered.
    SUPPORTED_DUNDER = {
        "__add__",
        "__radd__",
        "__sub__",
        "__rsub__",
        "__mul__",
        "__rmul__",
        "__truediv__",
        "__rtruediv__",
        "__floordiv__",
        "__rfloordiv__",
        "__pow__",
        "__rpow__",
        "__mod__",
        "__rmod__",
        "__round__",
        "__pos__",
        "__neg__",
        "__abs__",
        "__and__",
        "__rand__",
        "__or__",
        "__ror__",
        "__xor__",
        "__rxor__",
        "__invert__",
        "__lt__",
        "__le__",
        "__eq__",
        "__ne__",
        "__gt__",
        "__ge__",
    }

    # Set containing all operator names that have been defined across all registries.
    # Used for __dir__ method of SymbolicExpression
    ALL_REGISTERED_OPS = set()  # type: set[str]

    def __init__(self, name, super_registry=None):
        self.name = name
        self.super_registry = super_registry
        self.registered_ops = set()  # type: set[Operator]
        self.implementations = dict()  # type: dict[str, OperatorImplementationStore]
        self.check_super = dict()  # type: dict[str, bool]

    def register_op(self, operator: Operator, check_super=True):
        """
        :param operator: The operator to register.
        :param check_super: Bool indicating if the super register should be
            checked if no implementation for this operator can be found.
        """

        name = operator.name
        if name.startswith("__") and name.endswith("__"):
            if name not in OperatorRegistry.SUPPORTED_DUNDER:
                raise ValueError(f"Dunder method {name} is not supported.")

        if operator in self.registered_ops:
            raise ValueError(
                f"Operator {operator} ({name}) already registered in this operator"
                f" registry '{self.name}'"
            )
        if name in self.implementations:
            raise ValueError(
                f"Another operator with the name '{name}' has already been registered"
                " in this registry."
            )

        self.implementations[name] = OperatorImplementationStore(operator)
        self.check_super[name] = check_super

        self.registered_ops.add(operator)
        self.ALL_REGISTERED_OPS.add(name)

    def get_operator(self, name: str) -> Operator | None:
        if impl_store := self.implementations.get(name, None):
            return impl_store.operator
        return None

    def add_implementation(
        self,
        operator: Operator,
        impl: typing.Callable,
        signature: str,
        variant: str | None = None,
    ):
        if operator not in self.registered_ops:
            raise ValueError(
                f"Operator {operator} ({operator.name}) hasn't been registered in this"
                f" operator registry '{self.name}'"
            )

        signature = OperatorSignature.parse(signature)
        operator.validate_signature(signature)

        implementation_store = self.implementations[operator.name]
        op_impl = OperatorImpl(operator, impl, signature)

        if variant:
            implementation_store.add_variant(variant, op_impl)
        else:
            implementation_store.add_implementation(op_impl)

    def get_implementation(self, name, args_signature) -> TypedOperatorImpl:
        if name not in self.ALL_REGISTERED_OPS:
            raise ValueError(f"No operator named '{name}'.")

        if store := self.implementations.get(name):
            if impl := store.find_best_match(args_signature):
                return impl

        # If operation hasn't been defined in this registry, go to the parent
        # registry and check if it has been defined there.
        if self.super_registry is None or not self.check_super.get(name, True):
            raise ValueError(
                f"No implementation for operator '{name}' found that matches signature"
                f" '{args_signature}'."
            )
        return self.super_registry.get_implementation(name, args_signature)


class OperatorSignature:
    """
    Specification:

        signature ::= arguments "->" rtype
        arguments ::= (dtype ",")* terminal_arg
        terminal_arg ::= dtype | vararg
        vararg ::= dtype "..."
        rtype ::= dtype
        dtype ::= template | "int" | "float" | "str" | "bool" | and others...
        template ::= single uppercase character

    Examples:

        Function that takes two integers and returns an integer:
            int, int -> int

        Templated argument (templates consist of single uppercase characters):
            T, T -> T
            T, U -> bool

        Variable number of arguments:
            int... -> int

    """

    def __init__(self, args: tuple[str], rtype: str):
        """
        :param args: Tuple of argument types.
        :param rtype: The return type.
        """
        self.args = args
        self.rtype = rtype

    @classmethod
    def parse(cls, signature: str) -> OperatorSignature:
        def parse_cstypes(cst: str):
            # cstypes = comma seperated types
            types = cst.split(",")
            types = [t.strip() for t in types]
            types = [t for t in types if t]
            return types

        if "->" not in signature:
            raise ValueError("Invalid signature: arrow (->) missing.")

        arg_sig, r_sig = signature.split("->")
        args = parse_cstypes(arg_sig)
        rtype = parse_cstypes(r_sig)

        # Validate Signature
        if len(rtype) != 1:
            raise ValueError(
                f"Invalid operator signature '{signature}'. Expected exactly one return"
                " type."
            )

        rtype = rtype[0]

        base_args_types = [get_base_type(arg) for arg in args]
        base_rtype = get_base_type(rtype)

        for type in (*base_args_types, base_rtype):
            if not type.isalnum():
                raise ValueError(f"Invalid type '{type}'. Types must be alphanumeric.")

        # Validate Template
        # Output template must also occur in signature

        if is_template_type(base_rtype):
            if base_rtype not in base_args_types:
                raise ValueError(
                    f"Template return type '{base_rtype}' must also occur in the"
                    " argument signature."
                )

        # Validate vararg
        # Vararg can only occur for the very last element

        for arg in args[:-1]:
            if arg.endswith("..."):
                raise ValueError(f"Only last argument can be a vararg.")

        if rtype.endswith("..."):
            raise ValueError(f"Return value can't be a vararg.")

        return OperatorSignature(
            args=tuple(args),
            rtype=rtype,
        )

    def __repr__(self):
        return f"{ ', '.join(self.args)} -> {self.rtype}"

    def __hash__(self):
        return hash((self.args, self.rtype))

    def __eq__(self, other):
        if not isinstance(other, OperatorSignature):
            return False
        return self.args == other.args and self.rtype == other.rtype

    @property
    def is_vararg(self) -> bool:
        if len(self.args) == 0:
            return False
        return is_vararg_type(self.args[-1])


def is_template_type(s: str) -> bool:
    # Singe upper case ASCII character
    return len(s) == 1 and 65 <= ord(s) <= 90


def is_vararg_type(s: str) -> bool:
    return s.endswith("...")


def get_base_type(s: str) -> str:
    """Get base type from imput string without any decorators.
    -> Trims varargs ellipsis (...) from input.
    """
    if s.endswith("..."):
        return s[:-3]
    return s


class OperatorImplementationStore:
    """
    Stores all implementations for a specific operation in a trie according to
    their signature. This enables us to easily find the best matching
    operator implementation for a given set of input argument types.
    """

    @dataclasses.dataclass
    class TrieNode:
        __slots__ = ("value", "operator", "children", "is_vararg")
        value: str
        operator: OperatorImpl | None
        children: list[OperatorImplementationStore.TrieNode]
        is_vararg: bool

    def __init__(self, operator: Operator):
        self.operator = operator
        self.root = self.TrieNode("ROOT", None, [], False)

    def add_implementation(self, operator: OperatorImpl):
        node = self.get_node(operator.signature, create_missing=True)
        if node.operator is not None:
            raise ValueError(
                f"Implementation for signature {operator.signature} already defined."
            )

        node.operator = operator
        node.is_vararg |= operator.signature.is_vararg

    def add_variant(self, name: str, operator: OperatorImpl):
        node = self.get_node(operator.signature, create_missing=False)
        if node is None or node.operator is None:
            raise ValueError(
                f"No implementation for signature {operator.signature} found."
                " Make sure there is an exact match to add a variant."
            )

        assert node.operator.signature.rtype == operator.signature.rtype
        node.operator.add_variant(name, operator.impl)

    def get_node(self, signature: OperatorSignature, create_missing: bool = True):
        node = self.root
        for dtype in signature.args:
            dtype = get_base_type(dtype)
            for child in node.children:
                if dtype == child.value:
                    node = child
                    break
            else:
                if create_missing:
                    new_node = self.TrieNode(dtype, None, [], False)
                    node.children.append(new_node)
                    node = new_node
                else:
                    return None

        return node

    def find_best_match(self, signature: tuple[str]) -> TypedOperatorImpl | None:
        matches = list(self._find_matches(signature))

        if not matches:
            return None

        # Find best matching template.
        best_match = None
        best_score = (0xFFFF,)

        for match, templates in matches:
            if match.operator._precedence < best_score:
                best_match = (match, templates)
                best_score = match.operator._precedence

        rtype = best_match[0].operator.signature.rtype
        rtype = best_match[1].get(rtype, rtype)  # If it is a template -> Translate

        return TypedOperatorImpl.from_operator_impl(best_match[0].operator, rtype)

    def _find_matches(self, signature: tuple[str]):
        """Yield all operators that match the input signature"""

        # Case 0 arguments:
        if len(signature) == 0:
            yield (self.root, dict())
            return

        # Case 1+ args:

        def does_match(
            dtype: str,
            node: OperatorImplementationStore.TrieNode,
            templates: dict[str, str],
        ) -> bool:
            if is_template_type(node.value):
                t = templates.get(node.value)
                return t is None or dtype == t
            return dtype == node.value

        # Store tuple of (Node, index in signature, templates)
        stack = [(child, 0, dict()) for child in self.root.children]

        while stack:
            node, s_i, templates = stack.pop()
            dtype = signature[s_i]

            if not does_match(dtype, node, templates):
                continue

            if is_template_type(node.value) and node.value not in templates:
                # Insert dtype into templates dict
                templates = templates.copy()
                templates[node.value] = dtype

            if s_i + 1 == len(signature):
                if node.operator is not None:
                    yield (node, templates)
                continue

            children = iter(node.children)
            if node.is_vararg:
                children = itertools.chain(children, iter((node,)))

            for child in children:
                stack.append((child, s_i + 1, templates))


class OperatorRegistrationContextManager:
    def __init__(
        self,
        registry: OperatorRegistry,
        operator: Operator,
        # Options
        check_super=True,
    ):
        self.registry = registry
        self.operator = operator

        self.registry.register_op(operator, check_super=check_super)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, signature: str, variant: str = None):
        if not isinstance(signature, str):
            raise TypeError(f"Signature must be of type str.")

        def decorator(func):
            self.registry.add_implementation(self.operator, func, signature, variant)
            return func

        return decorator

    def auto(self, func=None, *, variant: str = None):
        if func is None:
            return partial(self.auto, variant=variant)

        if not self.operator.signatures:
            raise ValueError(f"Operator {self.operator} has not default signatures.")

        for sig in self.operator.signatures:
            self.registry.add_implementation(self.operator, func, sig, variant)

        return func

    def extension(self, extension: type[OperatorExtension], variant: str = None):
        if extension.operator != type(self.operator):
            raise ValueError(
                f"Operator extension for '{extension.operator.__name__}' can't "
                f"be applied to operator of type '{type(self.operator).__name__}'."
            )

        def decorator(func):
            for sig in extension.signatures:
                self.registry.add_implementation(self.operator, func, sig, variant)
            return func

        return decorator



from __future__ import annotations

from .core import ElementWise, OperatorExtension, Unary
from .numeric import Add, RAdd

__all__ = [
    "StringAdd",
    "StringRAdd",
    "Strip",
]


class StringAdd(OperatorExtension):
    operator = Add
    signatures = ["str, str -> str"]


class StringRAdd(OperatorExtension):
    operator = RAdd
    signatures = ["str, str -> str"]


####


class StringUnary(ElementWise, Unary):
    signatures = [
        "str -> str",
    ]


class Strip(StringUnary):
    name = "strip"



from __future__ import annotations

from .expressions import FunctionCall, SymbolicExpression

__all__ = [
    "count",
    "row_number",
]


def _sym_f_call(name, *args, **kwargs) -> SymbolicExpression[FunctionCall]:
    return SymbolicExpression(FunctionCall(name, *args, **kwargs))


def count(expr: SymbolicExpression = None):
    if expr is None:
        return _sym_f_call("count")
    else:
        return _sym_f_call("count", expr)


def row_number(*, arrange: list):
    return _sym_f_call("row_number", arrange=arrange)



from __future__ import annotations

from html import escape
from typing import Generic, TypeVar
from collections.abc import Iterable

from pydiverse.transform.core import verbs
from pydiverse.transform.core.column import Column, LambdaColumn
from pydiverse.transform.core.expressions import SymbolicExpression
from pydiverse.transform.core.table_impl import AbstractTableImpl

ImplT = TypeVar("ImplT", bound=AbstractTableImpl)


class Table(Generic[ImplT]):
    """
    All attributes of a table are columns except for the `_impl` attribute
    which is a reference to the underlying table implementation.
    """

    def __init__(self, implementation: ImplT):
        self._impl = implementation

    def __getitem__(self, key) -> SymbolicExpression[Column]:
        if isinstance(key, SymbolicExpression):
            key = key._
        return SymbolicExpression(self._impl.get_col(key))

    def __setitem__(self, col, expr):
        """Mutate a column
        :param col: Either a str or SymbolicColumn
        """
        col_name = None

        if isinstance(col, SymbolicExpression):
            underlying = col._
            if isinstance(underlying, (Column, LambdaColumn)):
                col_name = underlying.name
        elif isinstance(col, str):
            col_name = col

        if not col_name:
            raise KeyError(
                f"Invalid key {col}. Must be either a string, Column or LambdaColumn."
            )
        self._impl = (self >> verbs.mutate(**{col_name: expr}))._impl

    def __getattr__(self, name) -> SymbolicExpression[Column]:
        return SymbolicExpression(self._impl.get_col(name))

    def __iter__(self) -> Iterable[SymbolicExpression[Column]]:
        # Capture current state (this allows modifying the table inside a loop)
        cols = [
            SymbolicExpression(self._impl.get_col(name))
            for name, _ in self._impl.selected_cols()
        ]
        return iter(cols)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self._impl == other._impl

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._impl)

    def __dir__(self):
        return sorted(self._impl.named_cols.fwd.keys())

    def __contains__(self, item):
        if isinstance(item, SymbolicExpression):
            item = item._
        if isinstance(item, LambdaColumn):
            return item.name in self._impl.named_cols.fwd
        if isinstance(item, Column):
            return item.uuid in self._impl.available_cols
        return False

    def __copy__(self):
        impl_copy = self._impl.copy()
        return self.__class__(impl_copy)

    def __str__(self):
        try:
            return (
                f"Table: {self._impl.name}, backend: {type(self._impl).__name__}\n"
                f"{self >> verbs.collect()}"
            )
        except Exception as e:
            return (
                f"Table: {self._impl.name}, backend: {type(self._impl).__name__}\n"
                "Failed to collect table due to an exception:\n"
                f"{type(e).__name__}: {str(e)}"
            )

    def _repr_html_(self) -> str | None:
        html = (
            f"Table <code>{self._impl.name}</code> using"
            f" <code>{type(self._impl).__name__}</code> backend:</br>"
        )
        try:
            # TODO: For lazy backend only show preview (eg. take first 20 rows)
            html += (self >> verbs.collect())._repr_html_()
        except Exception as e:
            html += (
                "</br><pre>Failed to collect table due to an exception:\n"
                f"{escape(e.__class__.__name__)}: {escape(str(e))}</pre>"
            )
        return html

    def _repr_pretty_(self, p, cycle):
        p.text(str(self) if not cycle else "...")



from __future__ import annotations

import typing
from dataclasses import dataclass

from pydiverse.transform.core import column
from pydiverse.transform.core.expressions import expressions

__all__ = (
    "traverse",
    "sign_peeler",
    "OrderingDescriptor",
    "translate_ordering",
)

T = typing.TypeVar("T")


def traverse(obj: T, callback: typing.Callable) -> T:
    if isinstance(obj, list):
        return [traverse(elem, callback) for elem in obj]
    if isinstance(obj, dict):
        return {k: traverse(v, callback) for k, v in obj.items()}
    if isinstance(obj, tuple):
        if type(obj) != tuple:
            # Named tuples cause problems
            raise Exception
        return tuple(traverse(elem, callback) for elem in obj)

    return callback(obj)


def sign_peeler(expr):
    """Remove unary - and + prefix and return the sign
    :return: `True` for `+` and `False` for `-`
    """
    num_neg = 0
    while isinstance(expr, expressions.FunctionCall):
        if expr.name == "__neg__":
            num_neg += 1
            expr = expr.args[0]
        elif expr.name == "__pos__":
            expr = expr.args[0]
        else:
            break
    return expr, num_neg % 2 == 0


####


@dataclass
class OrderingDescriptor:
    __slots__ = ("order", "asc", "nulls_first")

    order: typing.Any
    asc: bool
    nulls_first: bool


def translate_ordering(tbl, order_list) -> list[OrderingDescriptor]:
    ordering = []
    for arg in order_list:
        col, ascending = sign_peeler(arg)
        if not isinstance(col, (column.Column, column.LambdaColumn)):
            raise ValueError(
                "Arguments to arrange must be of type 'Column' or 'LambdaColumn' and"
                f" not '{type(col)}'."
            )
        col = tbl.resolve_lambda_cols(col)
        ordering.append(OrderingDescriptor(col, ascending, False))

    return ordering
